""" Module for species and mixture """

import numpy as np

ATOMIC_MASS = {"C": 12.0107, "H": 1.00784, "O": 15.999, "N": 14.0067}


class SpeciesState:
    """
    *Class managing species*

    :Attributes:

        - **_name** - Name of the species
        - **_atoms** - Dict['CHON'] of atom numbers
        - **_molar_mass** - Molar mass of the species
        - **_mass_fraction** - Mass fraction of the species
        - **_stream** - Input stream of the species
        - **_mass_fraction_stream** - Mass fraction streamwise
    """

    def __init__(self, name, mass_fraction, stream=None, mass_fraction_stream=0.0):
        """Initialize SpeciesState class"""

        self._name = name.upper()
        self._atoms = self._get_atoms()
        self._molar_mass = self._get_molar_mass()
        self._mass_fraction = mass_fraction
        self._stream = stream
        self._mass_fraction_stream = mass_fraction_stream

    @property
    def atoms(self):
        """returns dict['CHON'] of number of atoms"""
        return self._atoms

    @property
    def molar_mass(self):
        """returns species molar mass"""
        return self._molar_mass

    @property
    def name(self):
        """returns species name"""
        return self._name

    def set_stream_data(self, stream, mass_frac):
        """
        *Set stream of the species and streamwise mass fraction*

        :param stream: Input stream number of the species (1 for fuel, 2 for oxydizer)
        :type stream: int
        :param mass_frac: Streamsize mass fraction of the species:
        :type mass_frac: float
        """
        self._stream = stream
        self._mass_fraction_stream = mass_frac

    def mass_fraction(self, stream=None):
        """
        *Returns either mass fraction or stream-wize mass fraction*

        :returns: Mass Fraction of the species
        """
        if stream is None:
            mass_fraction = self._mass_fraction
        else:
            mass_fraction = self._mass_fraction_stream * (stream == self._stream)
        return mass_fraction

    def mass(self):
        """
        *Compute the mass of the species*

        :returns: Mass of the species
        """
        mass = 0.0
        for atom, n_atom in self.atoms.items():
            mass += n_atom * ATOMIC_MASS[atom]
        return mass

    def _get_atoms(self):
        """
        *Get number of atoms of "CHON" in the species*

        :returns: Dict['CHON'] of number of atoms
        """
        atoms = {}
        idx = 0
        string_chon = self._name
        if "-" in string_chon:
            string_chon = self._name.split("-")[0]
        size = len(string_chon)
        while idx < size:
            atom = string_chon[idx]
            n_atom = ""
            idx += 1
            while idx < size and string_chon[idx] not in ATOMIC_MASS:
                n_atom += string_chon[idx]
                idx += 1
            if n_atom == "":
                n_atom = "1."
            atoms[atom] = float(n_atom)

        for atom in ATOMIC_MASS:
            if atom not in atoms:
                atoms[atom] = 0.0

        return atoms

    def _get_molar_mass(self):
        """
        *Compute the molar mass of the species*

            m_i = sum_j (n_i,j * M_j)

        with :

            - i : Species
            - j : Atom

        :returns: Molar mass of the species
        """

        molar_mass = 0.0
        for atom, n_atoms in self._atoms.items():
            molar_mass += n_atoms * ATOMIC_MASS[atom]

        return molar_mass


# TODO : awgh very complicated PI, we should do something easier next time!!!


class MixtureState:
    r"""\
    *Class managing mixture state*

    ::
              __ Stream 2
             /
        ----/---------------------
         -- m_ox --->
        -----------|
        -----------|
         -- m_fuel ->
        ---\----------------------
            \_ Stream 1

    :Attributes:

        - **_stream_dict** - Dict['O2', 'N2', ...] of dict['stream', 'mass_frac']
        - **_species** - List of SpeciesState object
        - **_mixture_fraction** - Mixture fraction of the mixture based on Bilger's
        - **_far** - Fuel Air Ratio of the mixture
        - **_far_st** - Stoechiometric Fuel Air Ratio of the mixture
        - **_phi** - Equivalence ratio of the mixture
    """

    def __init__(self, species_dict, fuel, stream_update=None, convert_sp=None):
        """Initialize MixtureState

        :param species_dict: dict of species mass fractions
            all KEYS must be CHON compositions
            accepted : O2, H2O, CH4, C12.34H21.07I68

        :param fuel: string, key of the fuel in species_dict

        :stream_update: - Dict['O2', 'N2', ...] of dict['stream', 'mass_frac']
            e.g.
            {
                "O2": {"stream": 2, "mass_frac": 0.2},
                "N2": {"stream": 2, "mass_frac": 0.7},
                "CH4": {"stream": 2, "mass_frac": 0.1},
                "O2": {"stream": 1, "mass_frac": 0.0},
                "N2": {"stream": 1, "mass_frac": 0.0},
                "CH4": {"stream": 1, "mass_frac": 1.0},
            }

        :param convert_sp: optionnal list of tuples , translates keys if needed


        Example of call:

        MixtureState(
            {O2: ..., N2: ..., KERO: ...},
            "KERO",
            convert_sp=(("KERO", "C10.45678H20.45678),)
        )


        """

        if stream_update is None:
            stream_update = {}

        if convert_sp is not None:
            for convert in convert_sp:
                species_dict[convert[1]] = species_dict[convert[0]]
                del species_dict[convert[0]]
                if convert[0] == fuel:
                    fuel = convert[1]

        self._stream_dict = {
            "O2": {"stream": 2, "mass_frac": 0.2331},
            "N2": {"stream": 2, "mass_frac": 0.7669},
            fuel: {"stream": 1, "mass_frac": 1.0},
        }

        self._species = self._instanciate_species_atomic(species_dict)
        self._fuel = self.species_by_name(fuel)
        self._check_mass_fraction()

        self._update_streams(stream_update)

        self._mixture_fraction = self._compute_mixture_fraction()
        self._far = self._compute_far()
        self._far_st = self._compute_far_st()
        self._phi = self._compute_equivalence_ratio()

    @property
    def species(self):
        """return list of SpeciesState object"""
        return self._species

    @property
    def species_name(self):
        """Return list of species names"""
        return [species.name for species in self._species]

    @property
    def mixture_fraction(self):
        """Return mixture fraction"""
        return self._mixture_fraction

    @property
    def far(self):
        """return Fuel Air Ratio"""
        return self._far

    @property
    def afr(self):
        """return Air Fuel Ratio"""
        return 1 / self._far

    @property
    def far_st(self):
        """return stoechiometric Fuel Air Ratio"""
        return self._far_st

    @property
    def equivalence_ratio(self):
        """returns equivalence ratio"""
        return self._phi

    def _instanciate_species_atomic(self, species_dict):
        # pylint: disable=no-self-use
        """
        *Instanciate SpeciesState object for each species of the mixture*

        :param species_dict: Dict['O2', 'N2', ..] of species mass fraction
        :param fuel: Fuel name
        :type fuel: str

        :returns: List of SpeciesState object
        """
        species_list = []
        for species, mass_fraction in species_dict.items():
            species_list.append(SpeciesState(species, mass_fraction))
        return species_list

    def _update_streams(self, stream_update):
        """
        *Update stream data for mixture's species*

        :param stream_update: Dict['O2', 'N2', ...] of dict['stream', 'mass_frac']
        """
        self._stream_dict.update(stream_update)

        for species, stream_data in self._stream_dict.items():
            self.species_by_name(species).set_stream_data(
                stream_data["stream"], stream_data["mass_frac"]
            )

    def species_by_name(self, name):
        """
        *Gets SpeciesState by name*

        :param name: Name of the species
        :type name: str

        :returns: SpeciesState object matching with name
        """
        try:
            idx = self.species_name.index(name)
        except ValueError:
            msg = "Species name '" + name + "' does not match with any species :"
            msg += "\n - "
            msg += "\n - ".join(self.species_name)
            raise NameError(msg)
        return self._species[idx]

    def _check_mass_fraction(self):
        """ """
        mass_fraction = 0.0
        for species in self._species:
            mass_fraction += species.mass_fraction()
        if not np.allclose(mass_fraction, 1.0, atol=1e-4):
            non_unit = np.where(np.abs(mass_fraction - 1.0) > 1e-4)[0]
            msg = "Sum of mass fractions is not unit in "
            msg += f"{len(non_unit)} / {len(mass_fraction)} of solutions.\n"
            raise RuntimeError(msg)

    def elem_mass_frac(self, atom, stream=None):
        """
        *Compute elemental mass fraction of atom j in mixture*

        For each species i, get the elemental mass fraction of the atom j.

                           a_i,j * M_j * Y_i
            Y_j = sum_i (---------------------)
                                  M_i
        with :

            - **a_i,j** : Number of atom j in species i
            - **M_j** : Molar mass of atom j
            - **M_i** : Molar mass of species i
            - **Y_i** : Mass fraction of species i

        If stream is not None, the mass_fraction is defined as the mass_fraction of \
        the species i in a the stream s :

            - s = 1 : Fuel stream
            - s = 2 : Oxydizer stream

                            a_i,j * M_j * Y_i,s
            Y_j,s = sum_i (---------------------)
                                   M_i
        with :

            - **Y_i,s** : Mass fraction of species i in stream s
        """

        el_mass_frac = 0.0
        for species in self._species:
            el_mass_frac += np.divide(
                species.atoms[atom] * ATOMIC_MASS[atom] * species.mass_fraction(stream),
                species.molar_mass,
            )

        return el_mass_frac

    def _compute_mixture_fraction(self):
        # pylint: disable=invalid-name
        """
        *Compute mixture fraction of the mixture*

        The mixture fraction Z defined by Bilger as :

              Y_C  / m.M_C  + Y_H  / n.M_H + (Y_O,2 - Y_O) / nu_O2.M_O
        Z = -----------------------------------------------------------
             Y_C,1 / m.M_C + Y_H,1 / n.M_H   +    Y_O,2    / nu_O2.M_O

        with :

            - **Y_j** : Elemental mass fraction of element j
            - **Y_j,s** : Elemental mass fraction of element j in stream s
            - **m, n** : Respectively number of carbon and hydrogen atoms in fuel
            - **M_j** : Molar mass of element j
            - **nu_O2** : Number of moles of O2
        """
        nu_o2 = np.add(self._fuel.atoms["C"], self._fuel.atoms["H"] / 4)

        num_C = np.divide(
            self.elem_mass_frac("C"), self._fuel.atoms["C"] * ATOMIC_MASS["C"]
        )
        num_H = np.divide(
            self.elem_mass_frac("H"), self._fuel.atoms["H"] * ATOMIC_MASS["H"]
        )
        num_O = np.divide(
            self.elem_mass_frac("O", 2) - self.elem_mass_frac("O"),
            nu_o2 * ATOMIC_MASS["O"],
        )
        num = num_C + num_H + num_O

        den_C = np.divide(
            self.elem_mass_frac("C", 1), self._fuel.atoms["C"] * ATOMIC_MASS["C"]
        )
        den_H = np.divide(
            self.elem_mass_frac("H", 1), self._fuel.atoms["H"] * ATOMIC_MASS["H"]
        )
        den_O = np.divide(self.elem_mass_frac("O", 2), nu_o2 * ATOMIC_MASS["O"])
        den = den_C + den_H + den_O

        mixture_fraction = np.divide(num, den)

        return mixture_fraction

    def _compute_far(self):
        """
        *Compute FAR (Fuel Air Ratio) from mixture fraction*

        FAR = Z / (1 - Z)
        """

        far = np.divide(self.mixture_fraction, 1.0 - self.mixture_fraction)
        return far

    def _compute_far_st(self):
        """
        *Compute stoechiometric FAR from stoechiometric quantities*

                sum_j(s=1) m_j    Y_O2,2 * (m*M_C + n*M_H)
        FAR_s = --------------- = ------------------------
                sum_j(s=2) m_j        M_O2 * (m + n/4)
        """

        fuel_mass = 1.0 * self._fuel.molar_mass
        nu_o2 = np.add(self._fuel.atoms["C"], self._fuel.atoms["H"] / 4)

        far_st = np.divide(
            self.species_by_name("O2").mass_fraction(2) * fuel_mass,
            self.species_by_name("O2").molar_mass * nu_o2,
        )

        return far_st

    def _compute_equivalence_ratio(self):
        """
        *Compute equivalence ratio phi*

        phi = FAR / FAR_st
        """

        return np.divide(self.far, self.far_st)



import pytest
import numpy as np
from ms_thermo.kero_prim2cons import kero_prim2cons


def test_kero_prim2cons():
    rho, rhoE, rhoY = kero_prim2cons(300, 101325, 1)
    target_value_rho = 1.1858159472682281  # 1.2320962372505444
    target_value_rhoE = 256190.8849042345  # 266054.68156187434
    np.testing.assert_allclose(target_value_rho, rho, rtol=10e-6)
    np.testing.assert_allclose(target_value_rhoE, rhoE, rtol=10e-6)



"""Module with functions helpers to  create gasout for CFD solvers"""

import yaml
import copy
import h5py
import numpy as np
from h5cross import hdfdict
import shutil
from importlib import resources
from ms_thermo.state import State


HEADLOG = """
   ### Gasout tool from ms_thermo package ###

    This is a demonstrator!
    It has not been thoroughy tested yet!

"""

__all__ = [
    "gasout_with_input",
    "gasout_dump_default_input",
    "alter_state",
    "spherical_tanh_mask",
    "directional_linear_mask",
    "fraction_z_mask",
    "load_mesh_and_solution",
    "save_data_for_avbp",
    "gasout_tool",
]


def alter_state(state, alpha, temp_new=None, press_new=None, y_new=None, verbose=False):
    """
    Apply gasout alterations to a State.

    :param state: State before alterations
    :type state: State
    :param alpha: Alteration mask with values from 0 (no alteration) to 1 (full alteration)
    :type alpha: ndarray or scalar
    :param temp_new: New temperature to apply, defaults to None
    :type temp_new: ndarray or scalar, optional
    :param press_new: New temperature to apply, defaults to None
    :type press_new: ndarray or scalar, optional
    :param y_new: New mass fractions to apply, defaults to None
    :type y_new: dict[str, ndarray or scalar], optional
    :param verbose: Verbosity, defaults to False
    :type verbose: bool, optional

    :returns: log, debug log
    :rtype: str
    """
    log = str()

    # Order may be important? According to commit d36c7348, Yk must be updated before T
    if y_new is not None:
        y_mix = dict()
        state_species = list(state.mass_fracs.keys())

        for spec in y_new:
            if spec not in state_species:
                msgerr = "\n\n Species mismatch:"
                msgerr += "\n" + str(spec) + " is not part of the mixture:\n"
                msgerr += "/".join(state_species)
                msgerr += "\nCheck your input file pretty please..."
                raise RuntimeWarning(msgerr)

        for spec in state.mass_fracs:
            value = 0
            if spec in y_new:
                value = y_new[spec]

            y_mix[spec] = state.mass_fracs[spec] + alpha * (
                value - state.mass_fracs[spec]
            )
        state.mass_fracs = y_mix

    if temp_new is not None:
        temp = np.array(state.temperature)
        state.temperature = temp + alpha * (temp_new - temp)

    if press_new is not None:
        state.pressure = state.pressure + alpha * (press_new - state.pressure)

    if verbose:
        footprint = 100.0 * np.sum(alpha) / alpha.shape[0]
        log += f"Footprint : {footprint: .3f} %"
    return log


def spherical_tanh_mask(coor, center, radius, delta):
    """
    Define a spherical mask. 0 inside the sphere, 1 outside, with a tanh transition.

    :param coor: Array of spatial coordinates (shape (n, ndim))
    :type coor: ndarray
    :param center: Array of sphere center coordinates (shape (ndim,))
    :type center: list, tuple or ndarray
    :param radius: Radius of the sphere
    :type radius: float
    :param delta: Transition thickness at the edge of the sphere
    :type delta: float

    :returns: alpha, alteration mask
    :rtype: ndarray
    """
    center = np.array(center)
    center_ndim = center.shape[0]
    coor_ndim = coor.shape[-1]
    if center_ndim != coor_ndim:
        msg_err = "\n\nDimension mismatch in the spherical center"
        msg_err += "\nCenter is " + str(center_ndim) + "D"
        msg_err += "\nCoords are " + str(coor_ndim) + "D"
        msg_err += "\nCheck your input file pretty please..."
        raise RuntimeWarning(msg_err)

    r_coor = np.linalg.norm(coor - center, axis=1)
    alpha = 0.5 * (1.0 - np.tanh((r_coor - radius) / (delta / 2)))
    return alpha


def directional_linear_mask(coor, axis, transition_start, transition_end):
    """
    Define a directional mask aligned with the `axis` coordinate axis.

    If `transition_end` > `transition_start`, the mask is 0 before `transition_start` and
    1 after `transition_end.
    If `transition_start` > `transition_end`, the mask is 1 before `transition_end` and
    0 after `transition_start.
    A linear transition is imposed between `transition_start` and `transition_end`.

    :param coor: Array of spatial coordinates (shape (n, ndim))
    :type coor: ndarray
    :param axis: Coordinate axis of the mask (0 x, 1 y, 2 z)
    :type axis: int
    :param transition_start: Start point of the linear mask
    :type transition_start: float
    :param transition_end: End point of the linear mask
    :type transition_end: float

    :returns: alpha, alteration mask
    :rtype: ndarray
    """
    direction = coor[:, axis]

    alpha = (direction - transition_start) / (transition_end - transition_start)
    alpha = np.clip(alpha, 0.0, 1.0)

    return alpha


def fraction_z_mask(
    state,
    specfuel,
    zmin,
    zmax,
    fuel_mass_fracs=None,
    oxyd_mass_fracs=None,
    atom_ref="C",
    verbose=False,
):
    """
    Compute a mask based on the mixture fraction Z.

    The mask is 1 between `zmin` and `zmax` and 0 outside.

    :param state: ms_thermo State object
    :type state: State
    :param specfuel: Fuel species name
    :type specfuel: str
    :param zmin: mask disabled (0) below this value
    :type zmin: float
    :param zmax: mask disabled (0) over this value
    :type zmax: float
    :param fuel_mass_fracs: Fuel mass fractions, defaults to composition at peak fuel concentration
    :type fuel_mass_fracs: dict, optional
    :param oxyd_mass_fracs: Oxydizer mass fractions, defaults to air
    :type oxyd_mass_fracs: dict, optional
    :param atom_ref: Reference atom, defaults to C
    :type atom_ref: str, optional
    :param verbose: Verbosity, defaults to False
    :type verbose: bool, optional

    :returns: alpha, alteration mask
    :rtype: ndarray
    """
    z_frac = state.compute_z_frac(
        specfuel,
        fuel_mass_fracs,
        oxyd_mass_fracs,
        atom_ref,
        verbose,
    )

    z_mid = 0.5 * (zmax + zmin)
    z_gap = 0.5 * (zmax - zmin)
    alpha = np.where(abs(z_frac - z_mid) < z_gap, 1.0, 0)
    return alpha


def gasout_with_input(coor, state, in_nob):
    """
    Update a State with gasout actions.

    :param coor: Array of spatial coordinates (shape (n, ndim))
    :type coor: ndarray
    :param state: ms_thermo State object
    :type state: State
    :param in_nob: Contents of the input file
    :type in_nob: dict

    :returns: Tuple containing the updated State and a log
    :rtype: tuple
    """

    log = str(HEADLOG)
    log += "\nInitial state"
    log += "\n ==========="
    log += "\n" + state.__repr__()

    for i, action in enumerate(in_nob["actions"]):

        log += "\n\nAction " + str(i) + ":" + action["type"]

        if action["type"] == "directional_linear_mask":
            reqd_keys = [
                "direction",
                "transition_start",
                "transition_stop",
            ]
            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            if action["direction"] == "x":
                axis = 0
            elif action["direction"] == "y":
                axis = 1
            elif action["direction"] == "z":
                axis = 2
            else:
                msgerr = "\n INPUT ERROR : " + "Direction must be one of x, y or z"
                raise RuntimeWarning(log + msgerr)

            alpha = directional_linear_mask(
                coor,
                axis=axis,
                transition_start=action["transition_start"],
                transition_stop=action["transition_stop"],
            )

        elif action["type"] == "spherical_tanh_mask":
            reqd_keys = [
                "center",
                "radius",
                "delta",
            ]
            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            alpha = spherical_tanh_mask(
                coor,
                center=action["center"],
                radius=action["radius"],
                delta=action["delta"],
            )

        elif action["type"] == "fraction_z_mask":
            reqd_keys = [
                "specfuel",
                "atom_ref",
                "oxyd_mass_fracs",
                "fuel_mass_fracs",
                "zmax",
                "zmin",
            ]

            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            alpha = fraction_z_mask(
                state,
                action["specfuel"],
                zmin=action["zmin"],
                zmax=action["zmax"],
                fuel_mass_fracs=action["fuel_mass_fracs"],
                oxyd_mass_fracs=action["oxyd_mass_fracs"],
                atom_ref=action["specfuel"],
                verbose=True,
            )

        if "new_temperature" not in action:
            action["new_temperature"] = None
        if "new_pressure" not in action:
            action["new_pressure"] = None
        if "new_yk" not in action:
            action["new_yk"] = None

        out = alter_state(
            state,
            alpha,
            temp_new=action["new_temperature"],
            press_new=action["new_pressure"],
            y_new=action["new_yk"],
            verbose=True,
        )
        log += "\n" + out

    log += "\n\n Final state"
    log += "\n ==========="

    log += "\n" + state.__repr__()

    return state, log


def gasout_dump_default_input(fname):
    """Dump the default gasout input file"""
    default_gasout_input = str(
        resources.files(__package__) / "INPUT/default_gasout_input.yml"
    )

    shutil.copyfile(default_gasout_input, fname)


def _check_action_params(reqd_keys, provided_keys):
    """
    Check gasout action keys

    :param reqd_keys: List of required keys
    :type reqd_keys: list[str]
    :param provided_keys: List of keys in the input file
    :type provided_keys: list[str]

    :returns: Log of the accepted and missing keys
    :rtype: str
    """
    log = str()
    for key in provided_keys:
        if key not in reqd_keys + ["type", "new_temperature", "new_pressure", "new_yk"]:
            log += "\n.   - key " + key + " not accepted"
    if log != "":
        log += "\n.  - Accepted keys"
        log += "\n.      >".join(reqd_keys)

    for key in reqd_keys:
        if key not in provided_keys:
            log += "\n key " + key + " is missing."
    return log


def load_mesh_and_solution(fname, mname):
    """
    Load a mesh and solution into HDF5 objects

    :param fname: Filename of the solution
    :type fname: str
    :param mname: Filename of the mesh
    :type mname: str
    """
    print("Reading solution in ", fname)
    with h5py.File(fname, "r") as fin:
        sol = hdfdict.load(fin, lazy=False)
    print("Reading mesh in ", mname)
    with h5py.File(mname, "r") as fin:
        mesh = hdfdict.load(fin, lazy=False)
    return mesh, sol


def build_data_from_avbp(mesh, sol):
    """
    Build coordinates and a State from an AVBP mesh and solution

    :param mesh: AVBP mesh
    :type mesh: HDF5 object
    :param sol: AVBP solution
    :type sol: HDF5 object

    :returns: Tuple of coordinates and State
    :rtype: tuple
    """
    state = State.from_cons(
        sol["GaseousPhase"]["rho"], sol["GaseousPhase"]["rhoE"], sol["RhoSpecies"]
    )

    x_coor = mesh["Coordinates"]["x"]
    y_coor = mesh["Coordinates"]["y"]
    try:
        z_coor = mesh["Coordinates"]["z"]
        coor = np.stack((x_coor, y_coor, z_coor), axis=-1)
    except KeyError:
        coor = np.stack((x_coor, y_coor), axis=-1)

    return coor, state


def save_data_for_avbp(state, sol, fname):
    """Update the full solution with the state parameters"""
    sol_new = copy.deepcopy(sol)
    sol_new["GaseousPhase"]["rho"] = state.rho
    sol_new["GaseousPhase"]["rhoE"] = state.rho * state.energy
    for spec in sol_new["RhoSpecies"]:
        sol_new["RhoSpecies"][spec] = state.rho * state.mass_fracs[spec]

    try:
        sol_new["Additionals"]["temperature"] = state.temperature
        sol_new["Additionals"]["pressure"] = state.pressure
    except KeyError:
        print("-Additionals- group is not part of the solution.")

    print("Saving solution in ", fname)
    with h5py.File(fname, "w") as fout:
        hdfdict.dump(sol_new, fout)


def gasout_tool(inputfile):
    """Main call"""

    with open(inputfile, "r") as fin:
        in_nob = yaml.load(fin, Loader=yaml.Loader)

    mesh, sol = load_mesh_and_solution(in_nob["inst_solut"], in_nob["mesh"])

    coor, state = build_data_from_avbp(mesh, sol)

    state_new, log = gasout_with_input(coor, state, in_nob)

    print(log)

    save_data_for_avbp(state_new, sol, in_nob["inst_solut_output"])



""" unit tests for pyavbp.tools """

# import pytest
import numpy as np

# from ms_thermo.state import State
from ms_thermo.species import build_thermo_from_avbp, build_thermo_from_cantera


"""
************
Examples for State.
************

State is an object handling the Internal state of a CFD fluid, namely:
- density
- Total Energy
- Mass Fractions

Limits of State object
======================
Velocity is NOT partof the state object and must be treated separately.
State refer to a point (shape 1) or a set of points (shape n).
The spatial aspects, i.e. the position of the points,
or the mesh, must be handled separately.
"""
import numpy as np
from ms_thermo.state import State  # , build_thermo_from_cantera, build_thermo_from_avbp


def how_to_state_single():
    """

    live Example
    ============
    In the following a state for a single point is created
    using the primitive variables T, P, Y_k.
    """
    # Prerequisites : - a thermodynamic database (a default AVBP one is used)
    # species = build_thermo_from_cantera("./chem.cti")           #for cantera species database
    # species = build_thermo_from_chemkin("./thermo.dat")         #for chemkin species database
    # species = build_thermo_from_avbp("./species_database.dat")  #for avbp species database

    # State object initialization
    # basic : mass fractions, temperature, pressure setup
    # the variables can be in arrays of the same size
    state = State(
        temperature=600.0,  # Kelvin, optional, deaults to 300.
        pressure=100000.0,  # Pa, optional, defaults to 101325.
        mass_fractions_dict={"N2": 1.0},
    )  # ,#optional, defaults to {'O2':0.2325,'N2':0.7675}
    # species_db=species)     # defaults to the avbp species_database.dat from pyavbp

    # this shows the most common value, the minima and maxima of the current state variables
    print(state)

    # reachable variables
    print(state.energy)  # in J
    print(state.rho)  # in Kg.m^-3
    print(state.mass_fracs)  # no unit
    print(state.temperature)  # in K
    print(state.pressure)  # in Pa

    # those can be updated directly from energy or rho:
    state.energy = 216038.00954532798
    state.rho = 1.171918080399414

    # or from primitives : temperature, pressure, mass_fractions:
    state.mass_fracs = {
        "O2": 0.2325,
        "N2": 0.7675,
    }  # recomputes rho and mass fracs, keeps energy
    state.pressure = 101325.0  # recomputes density, keeps mass fractions and energy
    state.temperature = 300.0  # recomputes energy and density, keeps mass fractions

    # this can be useful to update the primitives all at once, or to ensure a meaningful state:
    state.update_state()

    # useful methods that does not update the state variables
    print(state.list_species())  # list of the species present in the mixture
    print(state.mix_energy(600.0))  # gets the energy (J) value at given temperature (K)
    print(
        state.mix_enthalpy(600.0)
    )  # gets the enthalpy (J) value at given temperature (K)
    print(
        state.mix_molecular_weight()
    )  # computes the total molecular weight of the mixture
    return state


def how_to_state_hundred():
    """

    live Example
    ============
    In the following a state for a hundred point is created
    using the primitive variables T, P, Y_k.
    """
    state = State(
        temperature=600.0 * np.ones(100),  # optional
        pressure=100000.0 * np.ones(100),  # optional
        mass_fractions_dict={"N2": 1.0 * np.ones(100)},
    )  # optional
    # WARNING be careful with the variables shapes, the main ones are energy, rho and mass_fracs
    state.update_state(
        temperature=300.0 * np.ones(100),  # optional
        pressure=101325.0 * np.ones(100),  # optional
        mass_fracs={"O2": 1.0 * np.ones(100)},
    )  # optional

    # this recomputes mass fractions (thus stores density and mass fractions)
    # then temperature (thus stores energy and density)
    # then pressure (thus stores density)
    return state


def how_to_state_conservatives():
    """

    live Example
    ============
    In the following a state for a single point is created
    using the conservative variables rho, rho_e, rho_y
    """
    # state can also be initalized from conservatives : rho, rho_e, rho_y in that order
    rho = 1.171918080399414
    rho_e = rho * 216038.00954532798
    rho_y = {"O2": 0.2325 * rho, "N2": 0.7675 * rho}
    state_cons = State.from_cons(rho, rho_e, rho_y)
    print(state_cons.temperature)
    print(state_cons.energy)
    print(state_cons.mix_energy(300.0))
    return state_cons


def test_species(datadir):
    """
    Test of the different databases loaders
    """
    database_file = datadir.join("species_database.dat")
    species = build_thermo_from_avbp(database_file)
    assert species["CH4"].molecular_weight == 0.0160423
    assert species["CH4"].total_enthalpy(300.0) == 628962.1812333644
    assert species["CH4"].total_energy(300.0) == 473480.1119540216
    assert species["CH4"].c_p(300.0) == 2253.1058514053584
    assert species["CH4"].c_v(300.0) == 1734.83228714086
    assert species["CH4"].gamma(300.0) == 1.2987456298260704

    database_file = datadir.join("chem.cti")
    species = build_thermo_from_cantera(database_file)
    assert species["CH4"].molecular_weight == 0.016043
    assert species["CH4"].total_enthalpy(300.0) == 664568.8428129908
    assert species["CH4"].total_energy(300.0) == 509093.55764188815
    assert species["CH4"].c_p(300.0) == 2229.001829937566
    assert species["CH4"].c_v(300.0) == 1710.7508793672314
    assert species["CH4"].gamma(300.0) == 1.3029377081261675


def test_state():
    """
    Unit tests of State
    """

    temperature = 300.0
    pressure = 101325.0
    mass_fracs = {"O2": 0.2325, "N2": 0.7675}
    energy = 216038.00954532798
    rho = 1.171918080399414
    enthalpy = 273041.4890558396
    mol_weight = 0.028848789032908505
    c_p = 1012.4558080496037
    c_v = 724.2530696082889
    gamma = 1.3979309864674632
    csound = 347.658253944254
    representation = "\nCurrent primitive state of the mixture \n"
    representation += "\t\t| Most Common |    Min    |    Max \n"
    representation += "----------------------------------------------------\n"
    representation += "             rho| 1.17192e+00 | 1.172e+00 | 1.172e+00 \n"
    representation += "          energy| 2.16038e+05 | 2.160e+05 | 2.160e+05 \n"
    representation += "     temperature| 3.00000e+02 | 3.000e+02 | 3.000e+02 \n"
    representation += "        pressure| 1.01325e+05 | 1.013e+05 | 1.013e+05 \n"
    representation += "            Y_O2| 2.32500e-01 | 2.325e-01 | 2.325e-01 \n"
    representation += "            Y_N2| 7.67500e-01 | 7.675e-01 | 7.675e-01 \n"

    state = how_to_state_single()
    assert state.temperature == temperature
    assert state.pressure == pressure
    assert state.mass_fracs == mass_fracs
    assert state.rho == rho
    assert state.energy == energy
    assert state.c_p == c_p
    assert state.c_v == c_v
    assert state.gamma == gamma
    assert state.csound == csound

    assert repr(state) == representation
    assert state.mix_molecular_weight() == mol_weight
    state.update_state()
    assert state.mass_fracs == mass_fracs
    assert state.rho == rho
    assert state.energy == energy

    state.mass_fracs = mass_fracs
    assert state.mass_fracs == mass_fracs
    state.temperature = temperature
    assert state.temperature == temperature
    state.pressure = pressure
    assert state.pressure == pressure

    state = how_to_state_hundred()
    np.testing.assert_allclose(
        state.mix_enthalpy(temperature * np.ones(100)), enthalpy * np.ones(100)
    )

    cons_state = how_to_state_conservatives()

    assert cons_state.temperature == temperature
    assert cons_state.pressure == pressure
    assert cons_state.mass_fracs == mass_fracs
    assert cons_state.rho == rho
    assert cons_state.mix_energy(temperature) == energy



#!/usr/bin/env python
"""
cli.py
Command line interface for tools in ms_thermo
"""

import click
import ms_thermo


def add_version(f):
    """
    Add the version of the tool to the help heading.
    :param f: function to decorate
    :return: decorated function
    """
    doc = f.__doc__
    f.__doc__ = (
        "Package " + ms_thermo.__name__ + " v" + ms_thermo.__version__ + "\n\n" + doc
    )
    return f


@click.group()
@add_version
def main_cli():
    """---------------    MS_THERMO  --------------------

    You are now using the Command line interface of MS-Thermo,
    a Python3 helper for reactive multispecies computation, created at CERFACS (https://cerfacs.fr).

    This is a python package currently installed in your python environement.
    See the full documentation at : https://ms-thermo.readthedocs.io/en/latest/.
    """
    pass


@click.command(help="Renamed as kero-tadia", deprecated=True)
@click.argument("temperature", nargs=1)
@click.argument("pressure", nargs=1)
@click.argument("phi", nargs=1)
def tadia(temperature, pressure, phi):
    return


main_cli.add_command(tadia)


@click.command()
@click.argument("temperature", nargs=1)
@click.argument("pressure", nargs=1)
@click.argument("phi", nargs=1)
def kero_tadia(temperature, pressure, phi):
    """Adiabatic flame temperature for a kerosene-air mixture.

    The initial TEMPERATURE is in Kelvins.
    The initial PRESURE in in Pascals.
    The PHI equivalence ratio equals 1 at stoechiometry.

    The computation is done by interpolation in a precomputed 2S_KERO_BFER table.
    The system considered to build the table was a 1D flame at constant pressure.
    """
    from ms_thermo.tadia import tadia_table

    burnt_temperature, yk_ = tadia_table(
        float(temperature), float(pressure), float(phi)
    )
    print(
        "\nThe adiabatic flame temperature of a mix "
        + f"C10H22-air from tables is : {burnt_temperature:.2f} K."
    )
    print("\nSpecies     |    Mass fraction")
    print("------------------------------")
    for specie in yk_:
        print(f"{specie:12s}|       {yk_[specie]:.3f}")


main_cli.add_command(kero_tadia)


@click.command(help="Renamed as kero-prim2cons", deprecated=True)
@click.argument("temperature", nargs=1, type=float)
@click.argument("pressure", nargs=1, type=float)
@click.argument("phi", nargs=1, type=float)
def fresh_gas(temperature, pressure, phi):
    return


main_cli.add_command(fresh_gas)


@click.command()
@click.argument("temperature", nargs=1, type=float)
@click.argument("pressure", nargs=1, type=float)
@click.argument("phi", nargs=1, type=float)
def kero_prim2cons(temperature, pressure, phi):
    """Primitive to conservative variable conversion for a kerosene-air mixture.

    The initial TEMPERATURE is in Kelvins.
    The initial PRESURE in in Pascals.
    The PHI equivalence ratio equals 1 at stoechiometry.

    The computation is done using a 100K piecewise linear table of enthalpies.
    """
    from ms_thermo.kero_prim2cons import kero_prim2cons

    rho, rhoE, rhoyk = kero_prim2cons(temperature, pressure, phi)
    print(f"\nrho       |  {rho:.3f} kg/m3")
    print(f"rhoE      |  {rhoE:.3f} J.kg/m3")
    print("rhoYk     |")
    for specie in rhoyk:
        print(f" {specie:9s}|  {rhoyk[specie]:.3f} mol.kg/m3")
    print("------------------------------")
    print("Yk        |")
    for specie in rhoyk:
        print(f" {specie:9s}|  {rhoyk[specie]/rho:.3f} [-]")


main_cli.add_command(kero_prim2cons)


@click.command()
@click.argument("temperature", nargs=1, type=float)
@click.argument("pressure", nargs=1, type=float)
@click.argument("phi", nargs=1, type=float)
@click.argument("fuel_name", nargs=1)
@click.argument("c_x", nargs=1, type=float)
@click.argument("h_y", nargs=1, type=float)
@click.argument("path2cti", nargs=1)
def hp_equil(temperature, pressure, phi, fuel_name, c_x, h_y, path2cti):
    """HP equilibrium using Cantera.

    The initial TEMPERATURE is in Kelvins.
    The initial PRESURE in in Pascals.
    The PHI equivalence ratio equals 1 at stoechiometry.
    The FUEL_NAME is the name of the fuel in your system.
    The path2cti is the path to your .CTI file.

    The computation is done using the Cantera library
    installed on your environement (not included in this package)
    The system considered is an HP equilibrium.

    See http://www.cerfacs.fr/cantera/ for more information and some .CTI files

    """
    from ms_thermo.tadia import tadia_cantera

    burnt_temperature, yk_ = tadia_cantera(
        temperature, pressure, phi, c_x, h_y, fuel_name, path2cti
    )
    print(
        "\nThe adiabatic flame temperature of a mix "
        + f"{str(fuel_name)}-air from cantera is : {burnt_temperature:.2f} K."
    )
    print("\nSpecies     |    Mass fraction")
    print("------------------------------")
    nb_other_species = 0
    y_other_species = 0
    for specie in yk_:
        if yk_[specie] < 1e-3:
            nb_other_species += 1
            y_other_species += yk_[specie]
        else:
            print(f"{specie:12s}|       {yk_[specie]:.3f}")
    print(f"+ {nb_other_species} others |       {y_other_species:.3f}")


main_cli.add_command(hp_equil)


@click.command()
@click.argument("phi", nargs=1, type=float)
@click.argument("c_x", nargs=1, type=float)
@click.argument("h_y", nargs=1, type=float)
@click.argument("fuel_name", nargs=1)
def yk_from_phi(phi, c_x, h_y, fuel_name):
    """Mass fractions of a fuel-air mixture.

    The PHI equivalence ratio equals 1 at stoechiometry.
    C_X is the number of carbon atoms in your fuel.
    H_Y is the number of hydrogen atoms in your fuel
    FUEL_NAME is the name of fuel species.
    """
    from ms_thermo.yk_from_phi import yk_from_phi

    yk_ = yk_from_phi(float(c_x), float(h_y), float(phi), str(fuel_name))
    print("\nSpecies     |    Mass fraction")
    print("------------------------------")
    for specie in yk_:
        print(f"{specie:12s}|       {yk_[specie]:.3f}")


main_cli.add_command(yk_from_phi)


@click.command()
@click.option("--new", is_flag=True, help="prefill INPUTFILE with default values")
@click.argument("inputfile", nargs=1)
def gasout(inputfile, new):
    """Apply GASOUT actions to a mixture.

    A GASOUT action is a local alteration of the thermodynamical
    state (temperature, pressure or mixture) of a mixture.

    The INPUTFILE, in .yaml format, controls the list of actions.
    Create a new INPUTFILE with the --new flag to see all possibilities.
    """
    from ms_thermo.gasout import gasout_dump_default_input, gasout_tool

    if new:
        gasout_dump_default_input(inputfile)
    else:
        gasout_tool(inputfile)


main_cli.add_command(gasout)



from ms_thermo.flame_params import FlameTable
from ms_thermo.yk_from_phi import yk_from_phi
from importlib import resources


__all__ = ["tadia_cantera", "tadia_table"]


def tadia_table(
    t_fresh_gases, p_fresh_gases, phi, fuel=None, fuel_name="KERO", c_x=10, h_y=20
):
    """
    *Compute the adiabatic flame temperature of a premixed kero/air mixture from tables*

    :param t_fresh_gases: Temperature of the fresh gases [K]
    :type t_fresh_gases: float
    :param p_fresh_gases: Pressure of the fresh gases [Pa]
    :type p_fresh_gases: float
    :param phi: Equivalence ratio [-]
    :type phi: float
    :param fuel: path to flame table AVBP
    :type fuel: str
    :param fuel_name: Name of the fuel species
    :type fuel_name: str
    :param c_x: nb. of C atoms
    :type c_x: float
    :param h_y: nb. of H atoms
    :type h_y: float

    :returns:
        - **t_burnt_gases** - Temperature of the burnt gases
        - **yk_burnt** - Dict of mass fractions of burnt gases
    """

    case = FlameTable()

    mass_molar = {"C": 0.0120107, "H": 0.00100797, "O2": 0.0319988, "N2": 0.0280134}

    if fuel is None:
        fuel_table = str(resources.files(__package__) / "INPUT/2S_KERO_BFER.h5")
        c_x = 10.0
        h_y = 20.0
    else:
        fuel_table = fuel

    yk_fresh = yk_from_phi(phi, c_x, h_y, fuel_name)

    mass_mol_fuel = c_x * mass_molar["C"] + h_y * mass_molar["H"]
    coeff_o2 = c_x + (h_y / 4)
    stoechio = coeff_o2 * mass_molar["O2"] / mass_mol_fuel

    yk_burnt = dict()
    yk_burnt["N2"] = yk_fresh["N2"]

    if phi <= 1:
        yk_burnt[fuel_name] = 0.0
        yk_burnt["O2"] = yk_fresh["O2"] - stoechio * yk_fresh[fuel_name]
        yk_burnt["CO2"] = (
            c_x
            * yk_fresh[fuel_name]
            * (mass_molar["O2"] + mass_molar["C"])
            / mass_mol_fuel
        )
        yk_burnt["H2O"] = (
            h_y
            / 2
            * yk_fresh[fuel_name]
            * (0.5 * mass_molar["O2"] + 2 * mass_molar["H"])
            / mass_mol_fuel
        )
    else:
        yk_burnt[fuel_name] = yk_fresh[fuel_name] - yk_fresh["O2"] / stoechio
        yk_burnt["O2"] = 0.0
        yk_burnt["CO2"] = (
            (1 / (1 + h_y / (4 * c_x)))
            * yk_fresh["O2"]
            * (mass_molar["O2"] + mass_molar["C"])
            / mass_molar["O2"]
        )
        yk_burnt["H2O"] = (
            (1 / (1 / 2 + 2 * c_x / h_y))
            * yk_fresh["O2"]
            * (0.5 * mass_molar["O2"] + 2 * mass_molar["H"])
            / mass_molar["O2"]
        )

    case.read_table_hdf(fuel_table)
    case.get_params(phi, t_fresh_gases, p_fresh_gases)

    t_burnt_gases = case.interpolated_data_dict["TEMPERATURE_BURNT"][0]

    return t_burnt_gases, yk_burnt


def tadia_cantera(t_fresh_gases, p_fresh_gases, phi, c_x, h_y, fuel_name, cti_file):
    """
    *Compute the adiabatic flame temperature of a premixed fuel/air mixture from Cantera*

    :param t_fresh_gases: Temperature of the fresh gases [K]
    :type t_fresh_gases: float
    :param p_fresh_gases: Pressure of the fresh gases [Pa]
    :type p_fresh_gases: float
    :param phi: Equivalence ratio [-]
    :type phi: float
    :param c_x: nb. of C atoms
    :type c_x: float
    :param h_y: nb. of H atoms
    :type h_y: float
    :param fuel_name: Name of the fuel species
    :type fuel_name: string
    :param cti_file: Path to the cti file to consider
    :type cti_file: string

    :returns:
        - **t_burnt_gases** - Temperature of the burnt gases
        - **yk_burnt** - Dict of mass fractions of burnt gases

    .. note:: **Warning**: This function may not be available if you do not have cantera in your environment
    """
    try:
        import cantera as ct
    except ModuleNotFoundError:
        print("Cantera module is not installed.")
        print("See instructions on http://www.cerfacs.fr/cantera/installation.php")
        raise ModuleNotFoundError
    case = ct.Solution(cti_file)
    case.Y = yk_from_phi(phi, c_x, h_y, fuel_name)
    case.TP = t_fresh_gases, p_fresh_gases
    case.equilibrate("HP")
    t_burnt_gases = case.T
    yk_burnt = case.mass_fraction_dict()
    return t_burnt_gases, yk_burnt



""" Module testing SpeciesState and MixtureState """

import numpy as np
import pytest

from ms_thermo.mixture_state import SpeciesState, MixtureState

DIM1 = 5


def test_mixture_state():
    """test of SpeciesAtomic and MixtureAtomic classes"""

    ATOMIC_MASS = {"C": 12.0107, "H": 1.00784, "O": 15.999, "N": 14.0067}

    def compute_mass_frac(phi, fuel, n_c, n_h):
        m_fuel = max((phi - 1) * (n_c * ATOMIC_MASS["C"] + n_h * ATOMIC_MASS["H"]), 0.0)
        m_o2 = max((1 - phi) * (n_c + n_h / 4) * 2 * ATOMIC_MASS["O"], 0.0)
        m_n2 = (n_c + n_h / 4) * 3.76 * 2 * ATOMIC_MASS["N"]
        m_co2 = phi * n_c * (ATOMIC_MASS["C"] + 2 * ATOMIC_MASS["O"])
        m_h2o = phi * n_h / 2 * (2 * ATOMIC_MASS["H"] + ATOMIC_MASS["O"])

        msum = m_fuel + m_o2 + m_n2 + m_co2 + m_h2o
        species_dict = {}
        species_dict[fuel] = m_fuel / msum * np.ones(DIM1)
        species_dict["O2"] = m_o2 / msum * np.ones(DIM1)
        species_dict["N2"] = m_n2 / msum * np.ones(DIM1)
        species_dict["CO2"] = m_co2 / msum * np.ones(DIM1)
        species_dict["H2O"] = m_h2o / msum * np.ones(DIM1)

        return species_dict

    fuel = "C10H20"
    phi = 1.0
    species_dict = compute_mass_frac(phi, fuel, 10, 20)
    mixture = MixtureState(species_dict, fuel)

    assert len(mixture.species) == 5
    assert isinstance(mixture.species[0], SpeciesState)
    assert mixture.species_name == [fuel, "O2", "N2", "CO2", "H2O"]
    np.testing.assert_allclose(mixture.mixture_fraction, 0.063775, atol=1e-6)
    np.testing.assert_allclose(mixture.equivalence_ratio, phi)
    np.testing.assert_allclose(mixture.far, 0.06812, atol=1e-6)
    np.testing.assert_allclose(mixture.far_st, 0.06812, atol=1e-6)
    np.testing.assert_allclose(mixture.afr, 14.680006, atol=1e-6)

    phi = 0.5
    species_dict = compute_mass_frac(phi, fuel, 10, 20)
    mixture = MixtureState(species_dict, fuel)
    np.testing.assert_allclose(mixture.mixture_fraction, 0.03295, atol=1e-6)
    np.testing.assert_allclose(mixture.equivalence_ratio, phi, atol=2e-4, rtol=4e-4)
    np.testing.assert_allclose(mixture.far, 0.034073, atol=1e-6)
    np.testing.assert_allclose(mixture.far_st, 0.06812, atol=1e-6)
    np.testing.assert_allclose(mixture.afr, 29.348641, atol=1e-6)

    with pytest.raises(NameError):
        mixture.species_by_name("dummy")

    species_dict["N2"] /= 2.0
    with pytest.raises(RuntimeError):
        mixture = MixtureState(species_dict, fuel)



"""
This script calculate mass_fraction of species from a Phi
"""

__all__ = ["yk_from_phi", "phi_from_far"]


YO2_IN_AIR = 0.2314

MOLAR_MASSES = {"C": 0.0120107, "H": 0.00100797, "O2": 0.0319988, "N2": 0.0280134}


def far_stochio(c_x: float, h_y: float) -> float:
    """Return the fuel air ratio stoechiometric according to c_x h_y

    valid for AIR , not for PURE OXYGEN!!!

    BTW, the example in https://en.wikipedia.org/wiki/Air%E2%80%93fuel_ratio is for pure oxygen...
    """
    mass_mol_fuel = c_x * MOLAR_MASSES["C"] + h_y * MOLAR_MASSES["H"]
    coeff_o2 = c_x + (h_y / 4)
    coef_air = 1.0 / YO2_IN_AIR
    afr_sto = coef_air * (coeff_o2 * MOLAR_MASSES["O2"]) / mass_mol_fuel
    far_sto = 1.0 / afr_sto
    return far_sto


def phi_from_far(far: float, c_x: float, h_y: float) -> float:
    """
    *Return phi coefficient with the fuel air ratio coeff + fuel composition (Valid for AIR, Not  PURE OXYGEN!!)**

    :param far: the air-fuel ratio
    :type far: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float

    :returns:
        - **phi** - Equivalence ratio
    """
    return far / far_stochio(c_x, h_y)


def far_from_phi(phi: float, c_x: float, h_y: float) -> float:
    """
    *Return fuel air ratio coeff  with the phi coefficient + fuel composition (Valid for AIR, Not  PURE OXYGEN!!)*

    :param phi: eq. ratio coef
    :type far: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float

    :returns:
        - **far** - Equivalence ratio
    """
    return phi * far_stochio(c_x, h_y)


# use "fuel" here as default. Beware it important for pyavbp.
def yk_from_phi(phi: float, c_x: float, h_y: float, fuel_name: str = "fuel") -> float:
    """
    *Return the species mass fractions in a fresh fuel-air mixture (Valid for AIR, Not  PURE OXYGEN!!)*

    :param phi: equivalence ratio
    :type phi: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float
    :param fuel_name: Name of the fuel
    :type fuel_name: str

    :returns:
        - **y_k** - A dict of mass fractions

    """
    y_k = dict()
    if phi == 0.0:
        y_k[fuel_name] = 0.0
    else:
        afr_stochio = 1.0 / far_stochio(c_x, h_y)
        y_k[fuel_name] = 1.0 / (
            1.0
            + (1.0 + 3.76 * (MOLAR_MASSES["N2"] / MOLAR_MASSES["O2"]))
            * (afr_stochio / phi)
        )  # TNC Poinsot Veynante , pp 11-12

    y_air = 1 - y_k[fuel_name]
    y_k["N2"] = (1 - YO2_IN_AIR) * y_air
    y_k["O2"] = YO2_IN_AIR * y_air

    return y_k



from ms_thermo.state import State
from ms_thermo.yk_from_phi import yk_from_phi

__all__ = ["kero_prim2cons"]


def kero_prim2cons(temperature, pressure, phi):
    """
    Compute conservative variables from primitive variables in a kerosene-air mixture.

    :param temperature: the fresh gas temperature
    :type temperature: float
    :param pressure: pressure of the fresh gas
    :type pressure: float
    :param phi: equivalence ratio of the air-fuel mixture
    :type phi: float
    :param fuel: fuel
    :type fuel: string

    :returns:
        - **rho** - Density
        - **rhoE** - Conservative energy
        - **rhoyk** - Dict of conservative mass fractions

    """
    yk = yk_from_phi(
        phi, 10, 22, "KERO"
    )  # This can be adapted to any simple hydrocarbon fuel

    gas = State(None, temperature, pressure, yk)
    rho = gas.rho
    rhoE = gas.rho * gas.energy
    rhoyk = dict()
    for specie in gas._y_k.keys():
        rhoyk[specie] = gas.rho * gas._y_k[specie]
    return rho, rhoE, rhoyk



"""
``State`` is an object handling the internal state of a gaseous mixture, namely:

- Density
- Total energy
- Species mass fractions

Limitations of the ``State`` object
===================================
- Velocity is not a property of a ``State`` and must be treated separately.
- The spatial aspects, i.e. the position of the points, or the mesh, must be handled separately.

.. warning:: ``State`` variables are represented as structured arrays that must have the *same* shape

Initializing a ``State``
========================
A ``State`` object can be initialized in two ways:

- From the temperature, pressure, and species mass fractions :math:`(T, P, Y_k)` through the default constructor::

    state = State(T, P, Yk)

- From conservative variables :math:`(\\rho, \\rho E, \\rho Y_k)` through the ``from_cons`` constructor::

    state = State.from_cons(rho, rhoE, rhoYk)

The constructor arguments :math:`(T, P, Y_k)` or :math:`(\\rho, \\rho E, \\rho Y_k)` can be scalars or multidimensional arrays.

.. warning::
    When initializing from conservative variables, :math:`T` is determined by a
    Newton-Raphson method to ensure that the mixture energy matches the input total energy.
    This is an **expensive** step that may take a long time for large inputs.

Transforming a ``State``
========================
After a ``State`` has been initialized, :math:`T`, :math:`P` and :math:`Y_k` can
independently be set to new values (`e.g.` ``myState.temperature = newTemperature``) and the other state variables are modified
accordingly:

- When setting a new value for :math:`T`, the other state variables are modified assuming an **isobaric and iso-composition** transformation from the previous state.
- When setting a new value for :math:`P`, the other state variables are modified assuming an **isothermal and iso-composition** transformation from the previous state.
- When setting a new value for :math:`Y_k`, the other state variables are modified assuming an **isothermal and isobaric** transformation from the previous state.

State transformations always satisfy the perfect gas equation of state

.. math:: P = \\rho \\frac{R}{W_{mix}} T
"""

from importlib import resources
import numpy as np
from scipy.optimize import newton
from scipy.stats import mode
from ms_thermo.species import build_thermo_from_avbp
from ms_thermo.constants import GAS_CST, ATOMIC_WEIGHTS
import warnings

__all__ = ["State"]


class State:
    """Container class to handle mixtures."""

    def __init__(
        self,
        species_db=None,
        temperature=300.0,
        pressure=101325.0,
        mass_fractions_dict=None,
    ):
        """
        Initialize a ``State`` object.

        :param species_db: Species database, defaults to AVBP species database
        :type species_db: Species, optional
        :param temperature: Temperature, defaults to 300 K
        :type temperature: ndarray or scalar, optional
        :param pressure: Pressure, defaults to 1 atm
        :type pressure: ndarray or scalar, optional
        :param mass_fractions_dict: Species mass fractions, defaults to air
        :type mass_fractions_dict: dict[str, ndarray or scalar], optional
        """
        self.species = species_db
        if self.species is None:
            self._default_thermo_database()

        self._y_k = dict()
        if mass_fractions_dict is None:
            self._y_k = {"O2": 0.2325, "N2": 0.7675}
        else:
            self._y_k = mass_fractions_dict
        self._check_y_input(self._y_k)

        self.rho = pressure * self.mix_w / (GAS_CST * temperature)

        self.energy = self.mix_energy(temperature)
        self._temperature = temperature

    def __repr__(self):
        """Define the value returned by calling State representation (printing State)"""
        mini = dict()
        maxi = dict()
        moco = dict()
        moco["rho"] = mode(self.rho, axis=None, keepdims=False)[0]
        mini["rho"] = np.min(self.rho)
        maxi["rho"] = np.max(self.rho)
        moco["energy"] = mode(self.energy, axis=None, keepdims=False)[0]
        mini["energy"] = np.min(self.energy)
        maxi["energy"] = np.max(self.energy)
        moco["temperature"] = mode(self.temperature, axis=None, keepdims=False)[0]
        mini["temperature"] = np.min(self.temperature)
        maxi["temperature"] = np.max(self.temperature)
        moco["pressure"] = mode(self.pressure, axis=None, keepdims=False)[0]
        mini["pressure"] = np.min(self.pressure)
        maxi["pressure"] = np.max(self.pressure)
        for specy in self._y_k:
            frac_specy = "Y_" + specy
            moco[frac_specy] = mode(self._y_k[specy], axis=None, keepdims=False)[0]
            mini[frac_specy] = np.min(self._y_k[specy])
            maxi[frac_specy] = np.max(self._y_k[specy])
        repr_str = "\nCurrent primitive state of the mixture \n"
        repr_str += "\t\t| Most Common |    Min    |    Max \n"
        repr_str += "-" * 52 + "\n"
        for var in moco:
            infos = (var, moco[var], mini[var], maxi[var])
            repr_str += "%16s| %#.05e | %#.03e | %#.03e \n" % infos

        return repr_str

    @classmethod
    def from_cons(cls, rho, rho_e, rho_y, species_db=None):
        """
        Class constructor from conservative values.

        :param rho: Density
        :type rho: ndarray or scalar
        :param rho_e: Total energy (conservative form)
        :type rho_e: ndarray or scalar
        :param rho_y: Species mass fractions (conservative form)
        :type rho_y: dict[str, ndarray or scalar]
        :param species_db: Species database, defaults to AVBP species database
        :type species_db: Species, optional
        """
        y_k = dict()
        for specy in rho_y:
            y_k[specy] = np.divide(rho_y[specy], rho)
        # State is initialized with default T, P
        state = cls(species_db=species_db, mass_fractions_dict=y_k)
        state.energy = np.divide(rho_e, rho)

        # Only rho and T need to be set. P is always derived from the equation of state
        state.rho = rho
        # Don't use the temperature setter as it modifies rho assuming an isobaric
        # transformation, which is not the case when starting from the default T,P
        state._temperature = state._energy_to_temperature(state.energy)
        return state

    @property
    def mass_fracs(self):
        """Getter or setter for the species mass fractions. Setting the mass fractions
        will modify the density assuming an isothermal and isobaric transformation.

        :rtype: dict[str, ndarray or scalar]
        """
        return self._y_k

    @mass_fracs.setter
    def mass_fracs(self, mass_fractions_dict):
        mol_weight_old = self.mix_w
        # actual update
        self._check_y_input(mass_fractions_dict)
        self._y_k = dict()
        for k in mass_fractions_dict:
            self._y_k[k] = mass_fractions_dict[k]

        mol_weight_new = self.mix_w
        self.rho = self.rho * mol_weight_new / mol_weight_old
        self.energy = self.mix_energy(self._temperature)

    @property
    def temperature(self):
        """Getter or setter for the temperature. Setting the temperature will recompute
        the total energy and modify the density assuming an isobaric transformation.

        :rtype: ndarray or scalar
        """
        return self._temperature

    @temperature.setter
    def temperature(self, temp):
        self.rho = np.divide(np.multiply(self.rho, self.temperature), temp)
        self._temperature = temp
        self.energy = self.mix_energy(temp)

    @property
    def pressure(self):
        """Getter or setter for the pressure. Setting the pressure will modify the
        density assuming an isothermal transformation.

        :rtype: ndarray or scalar
        """
        return self.rho * self.temperature * GAS_CST / self.mix_w

    @pressure.setter
    def pressure(self, press):
        self.rho = np.divide(np.multiply(self.rho, press), self.pressure)

    @property
    def gamma(self):
        """
        Get the heat capacity ratio

        :returns: **gamma** - Heat capacity ratio
        :rtype: ndarray or scalar
        """
        return self.c_p / self.c_v

    @property
    def c_p(self):
        r"""
        Get the mixture-averaged heat capacity at constant pressure

        .. warning::
            :math:`C_p` is computed like in AVBP as :math:`C_v+P/(\rho T)`, *not* as
            the weighted average of species :math:`C_{p,k}`

        :returns: **Cp** - Heat capacity at constant pressure
        :rtype: ndarray or scalar
        """
        # YOu thought it would this, me too
        # But  no!!!
        # c_p =  sum(
        #     self._y_k[k] * self.species[k].c_p(self.temperature)
        #     for k in self.list_spec

        c_p = self.c_v + self.pressure / (self.rho * self.temperature)

        return c_p

    @property
    def c_v(self):
        """
        Get the mixture-averaged heat capacity at constant volume

        :returns: **Cv** - Heat capacity at constant volume
        :rtype: ndarray or scalar
        """
        temp = self.temperature
        return sum(self._y_k[k] * self.species[k].c_v(temp) for k in self.list_spec)

    @property
    def csound(self):
        """
        Get the speed of sound

        :returns: **csound** - Speed of sound
        :rtype: ndarray or scalar
        """
        csound = np.sqrt(self.gamma * GAS_CST / self.mix_w * self.temperature)
        return csound

    @property
    def list_spec(self):
        """
        Get the names of the species

        :returns: **species_names** - List of species names
        :rtype: list[str]
        """
        return list(self._y_k.keys())

    @property
    def mix_w(self):
        r"""
        Compute the mixture molecular weight:

        .. math:: W_{mix} = \left[ \sum_{k=1}^{N_{sp}} \frac{Y_k}{W_k} \right]^{-1}

        :returns: **mix_mw** - Mixture molecular weight
        :rtype: ndarray or scalar
        """
        y_ov_w = [
            np.divide(self._y_k[k], self.species[k].molecular_weight)
            for k in self.list_spec
        ]
        mix_mw = 1.0 / np.sum(y_ov_w, axis=0)
        return mix_mw

    def mach(self, velocity):
        """
        Compute the Mach number

        :param velocity: Velocity
        :type velocity: ndarray or scalar

        :returns: **M** - Mach number
        :rtype: ndarray or scalar
        """
        return velocity / self.csound

    def temperature_total(self, velocity):
        r"""
        Compute the total temperature:

        .. math:: T_t = T \left[1+\frac{\gamma-1}{2}M^2 \right]

        where :math:`M` is the Mach number derived from the input velocity.
        This assumes an isentropic flow and constant gamma.

        :param velocity: Velocity
        :type velocity: ndarray or scalar

        :returns: **temp_total** - Total temperature
        :rtype: ndarray or scalar
        """
        return (
            1 + 0.5 * (self.gamma - 1) * self.mach(velocity) ** 2
        ) * self.temperature

    def pressure_total(self, velocity):
        r"""
        Compute the total pressure:

        .. math:: P_t = P \left[1+\frac{\gamma-1}{2}M^2 \right]^{\frac{\gamma}{\gamma-1}}

        where :math:`M` is the Mach number derived from the input velocity.
        This assumes an isentropic flow and constant gamma.

        :param velocity: Velocity
        :type velocity: ndarray or scalar
        :returns: **press_total** - Total pressure
        :rtype: ndarray or scalar
        """
        return (
            np.power(
                1 + 0.5 * (self.gamma - 1) * self.mach(velocity) ** 2,
                (self.gamma / (self.gamma - 1)),
            )
            * self.pressure
        )

    def mix_energy(self, temperature):
        r"""
        Compute the mixture total energy:

        .. math:: e = \sum_{k=1}^{N_{sp}} Y_k e_k

        :param temperature: Temperature
        :type temperature: ndarray or scalar

        :returns: **mix_energy** - Mixture total energy
        :rtype: ndarray or scalar
        """
        # Chesterton fence
        # func is called within setter self.temperature
        # do not use self.temperature here
        return sum(
            self._y_k[k] * self.species[k].total_energy(temperature)
            for k in self.list_spec
        )

    def mix_enthalpy(self, temperature):
        r"""
        Get mixture total enthalpy:

        .. math:: h = \sum_{k=1}^{N_{sp}} Y_k h_k

        :param temperature: Temperature
        :type temperature: ndarray or scalar

        :returns: **mix_enthalpy** - Mixture total enthalpy
        :rtype: ndarray or scalar
        """
        return sum(
            self._y_k[k] * self.species[k].total_enthalpy(temperature)
            for k in self.list_spec
        )

    def update_state(self, temperature=None, pressure=None, mass_fracs=None):
        """
        Compute density from temperature, pressure and mass fractions by assuming the
        following transformations:

            1) Isobaric and isothermal transformation,
            i.e (*P=cst*, *T=cst* and only **composition** is varying)

            2) Isobaric and iso-composition transformation,
            i.e (*P=cst*, *Y=cst* and only **temperature** is varying)

            3) Isothermal and iso-composition transformation,
            i.e (*T=cst*, *Y=cst* and only **pressure** is varying)

        :param temperature: Temperature to set, defaults to None
        :type temperature: ndarray or scalar, optional
        :param pressure: Pressure to set, defaults to None
        :type pressure: ndarray or scalar, optional
        :param mass_fracs: Mass fractions to set, defaults to None
        :type mass_fracs: dict[str, ndarray or scalar], optional
        """
        if mass_fracs is not None:
            self.mass_fracs = mass_fracs
        if temperature is not None:
            self.temperature = temperature
        if pressure is not None:
            self.pressure = pressure

    def compute_z_frac(
        self,
        specfuel,
        fuel_mass_fracs=None,
        oxyd_mass_fracs=None,
        atom_ref="C",
        verbose=False,
    ):
        """Compute the Z mixture fraction.

        0 oxidizer, 1 fuel

        :param specfuel: Fuel species
        :type specfuel: str
        :param fuel_mass_fracs: Fuel mass fractions, defaults to composition at peak fuel concentration
        :type fuel_mass_fracs: dict, optional
        :param oxyd_mass_fracs: Oxydizer mass fractions, defaults to air
        :type oxyd_mass_fracs: dict, optional
        :param atom_ref: Reference atom, defaults to C
        :type atom_ref: str, optional
        :param verbose: Verbosity, defaults to False
        :type verbose: bool, optional

        :returns: Z Mixture fraction
        :rtype: ndarray or scalar
        """
        log = "Computing Z fraction"
        if fuel_mass_fracs is None:
            # find where the maximum fuel concentration
            fuel_mass_fracs = dict()
            idx = np.argmax(self.mass_fracs[specfuel].ravel())
            for spec in self.mass_fracs:
                fuel_mass_fracs[spec] = self.mass_fracs[spec].ravel()[idx]
            log += "   Fuel mixture taken as : "
            for spec in fuel_mass_fracs:
                log += "\n.  - " + spec + ":" + str(fuel_mass_fracs[spec])
            log += "\n"
        if oxyd_mass_fracs is None:
            oxyd_mass_fracs = {"O2": 0.233, "N2": 0.767}
            log += "\n.  Oxidizer mixture taken as AIR."

        z_frac = 0.0
        z_fuel = 0.0
        z_oxyd = 0.0

        for spec in self.mass_fracs:
            nb_atoms = self.species[spec].atoms

            rel_weight = (
                nb_atoms[atom_ref]
                * ATOMIC_WEIGHTS[atom_ref.lower()]
                / self.species[spec].molecular_weight
            )
            z_frac += self.mass_fracs[spec] * rel_weight
            if spec in fuel_mass_fracs:
                z_fuel += fuel_mass_fracs[spec] * rel_weight
            if spec in oxyd_mass_fracs:
                z_oxyd += oxyd_mass_fracs[spec] * rel_weight

        z_frac = (z_frac - z_oxyd) / (z_fuel - z_oxyd)

        if verbose:
            print(log)
        return z_frac

    def _temp_energy_res(self, temperature):
        """Compute energy residuals from temperature.

        :param temperature: Temperature
        :type temperature: ndarray or scalar
        :returns: Energy residual
        :rtype: ndarray or scalar
        """
        comp_energy = self.mix_energy(temperature)
        total_energy = self.energy
        diff = 1.0 - comp_energy / total_energy
        return diff

    def _default_thermo_database(self):
        """
        Initialize the thermodynamic database if none is provided.

        By default it is the AVBP standard species database.
        """

        thermo_db_file = str(
            resources.files(__package__) / "INPUT/species_database.dat"
        )

        self.species = build_thermo_from_avbp(thermo_db_file)

    def _check_y_input(self, mass_fractions_dict):
        """
        Check if the names of the species are in the database
        the sum of mixture species mass fractions is unity

        :param mass_fractions_dict: Mass fractions
        :type mass_fractions_dict: dict[str, ndarray or scalar]
        """
        y_tol = 1.0e-5
        not_one_msg = "Mass fraction sum is not unity at point %d"
        not_in_db_msg = "Species %s is not present in the database"

        # check species presence
        for specie in mass_fractions_dict:
            if specie not in self.species:
                msg = not_in_db_msg % specie
                raise ValueError(msg)

        # check sum
        sum_y = 0
        for specie in mass_fractions_dict:
            sum_y += mass_fractions_dict[specie]
        conds = np.abs(sum_y - 1.0) > y_tol
        is_not_one = np.any(conds)
        if is_not_one:
            index = np.where(conds)[0][0]
            msg = not_one_msg % index
            msg += "\n min:" + str(np.min(sum_y))
            msg += "\n max:" + str(np.max(sum_y))
            raise ValueError(msg)

    def _energy_to_temperature(self, energy):
        r"""
        Find the temperature `T` that satisfies the constraint:

        .. math:: e = \sum_{k=1}^{N_{sp}} Y_k e_k(T)

        given a target value of the total energy `e`.

        :param energy: Target total energy
        :type energy: ndarray or scalar

        :returns: Temperature
        :rtype: ndarray or scalar
        """
        guess = np.ones_like(energy) * 300.0
        temperature = newton(self._temp_energy_res, guess, tol=1e-12)
        return temperature

    # Deprecated ====================================================================

    def list_species(self):
        r"""
        *Return primitives species names*.

        :returns: **species_names** - A list( ) of primitives species names

        """
        warnings.simplefilter("ignore")
        msgwarn = "ms_thermo.state.list_species() is deprecated,"
        msgwarn += "use instead ms_thermo.state.list_spec"
        warnings.warn(msgwarn)
        print(msgwarn)
        return self.list_spec

    def mix_molecular_weight(self):
        r"""
        *Compute mixture molecular weight following the formula :*

        .. math:: W_{mix} = \left[ \sum_{k=1}^{N_{sp}} \frac{Y_k}{W_k} \right]^{-1}

        :returns: **mix_mw** (float) - Mixture molecular weight
        """
        warnings.simplefilter("ignore")
        msgwarn = "ms_thermo.state.mix_molecular_weight() is deprecated,"
        msgwarn += "use instead ms_thermo.state.mix_w"
        warnings.warn(msgwarn)
        print(msgwarn)
        return self.mix_w



import numpy as np
from ms_thermo.tadia import tadia_table


def test_tadia_table():

    burned_t, yburnt = tadia_table(300, 101325, 1)
    target_value = 2312.990462472097

    np.testing.assert_allclose(target_value, burned_t, rtol=10e-6)
    print([k for k in yburnt.values()])
    target_value = [
        0.7566773271356266,
        0.0,
        0.17472881177111413,
        0.048670525063567914,
        0.019923336029691263,
    ]
    np.testing.assert_allclose(target_value, [k for k in yburnt.values()], rtol=10e-6)



import pytest
from ms_thermo.yk_from_phi import yk_from_phi, far_from_phi, phi_from_far, far_stochio


def test_far_stochio():
    # test on kerosen
    cx = 9.7396
    hy = 20.0542
    afr_sr_ref = 14.870368526488049
    # afr_st = 1./far_st_ref
    # assert afr_st==afr_sr_ref  # AFR of Kerozen is around 15
    far_st_ref = 1.0 / afr_sr_ref

    far_st = far_stochio(cx, hy)
    assert far_st == far_st_ref


def test_far_from_phi():
    # test from SAE : FAR 102.57 /1000   <> Phi 1.51
    cx = 9.7396
    hy = 20.0542
    phi_ref = 1.51
    far = far_from_phi(phi_ref, cx, hy)
    assert far == 0.10154422180662784


def test_phi_from_far():
    # test from SAE : FAR 102.57 /1000   <> Phi 1.51
    cx = 9.7396
    hy = 20.0542
    far_ref = 0.10257
    phi = phi_from_far(far_ref, cx, hy)
    assert phi == 1.525253699761879


def test_ykfromphi():
    yf_methane = yk_from_phi(1.0, 1, 4, "CH4")["CH4"]
    yf_c3h8 = yk_from_phi(1.0, 3, 8, "C3H8")["C3H8"]
    yf_c8h18 = yk_from_phi(1.0, 8, 18, "C8H18")["C8H18"]
    yf_h2 = yk_from_phi(1.0, 0, 2, "H2")["H2"]

    assert pytest.approx(yf_methane, 0.00001) == 0.01333564942487222
    assert pytest.approx(yf_c3h8, 0.00001) == 0.014642729952608102
    assert pytest.approx(yf_c8h18, 0.00001) == 0.015164593008774152
    assert pytest.approx(yf_h2, 0.00001) == 0.006747884056174038
